#!/usr/bin/ruby
# ==============================================================================
#
# name          :   bdiff
# project no.   :   1502007
# description   :   Agregates a diff over two parts lists or generates a total
#                   parts list over 2+ files.
# created       :   2016-07-12
$AUTHOR = "Gerald Mayer"
$VERSION = "0.2.1"
# notes         :   Bei Diff mit Drag'n Drop: die beiden Markierten Listen
#                   (alt+neu) mit dem Mauszeiger über der neuen Datei packen.
#                   In Windows wird die "gepackte" Datei das erste Argument.
# changes:
#   2018-01-31 v0.2.1 G. Mayer: Bugfix: hardcoded Arguments korrigiert
#                               Bugfix: Probleme bei neuen Listeneinträgen (Artikeln) korrigiert.
#   2016-09-21 v0.2.0 G. Mayer: Programmname auf 'bdiff.rb' geändert
#
# ==============================================================================
require 'pathname'



# ==============================================================================
#  global variables
#
$ARGSMIN = 2        # Mindestens zwei Argumente müssen angegeben werden
$POSID = 0          # default Position (Spalte) der Artikelnummer (ID)
$POSQTY = 2         # default Position (Spalte) der Bestellmenge
$SEP = ";"          # default Separator für die Spalten
$TEXTQUAL = "\""    # Textqualifizierer (Texterkennungszeichen)
$MODE = "1"         # default Modus: 0=Summenstk, 1=Mehrung, 2=Minderung
$POSTFIX0 = "_total"
$POSTFIX1 = "_increases"
$POSTFIX2 = "_reductions"




# ==============================================================================
# description  : Main function
# requires     : -
# returns      : -
def main()
  # Überprüfen der übergebenen Argumente:
  x = checkArgs(ARGV)
  outErr(x) if x != "0"

  # Initialisierung:
  argList = init(ARGV)

  # Dateiinhalte lesen:
  content = getContent(argList["files"])

  # Start Diff:
  newContent = diff(argList, content)

  # Neuen Inhalt (Differenz- oder Summenstückliste) in Zieldatei schreiben:
  destFile = argList["destFile"]
  f = File.new(destFile, "w+")
  f.write(newContent)
  f.close
end




# ==============================================================================
# description  : Überprüfung der übergebenen Argumente. Entsprechen diese den
#                Anforderungen?
# requires     : -
# returns      : errCode (Error code) as Integer
# args         : an Skript übergebene Argumente as Array
def checkArgs(args)
  # Error: To less arguments:
  return "1" if args.size < 1

  # Nur ein Argument übergeben:
  if args.size == 1 then
    case args.at(0)
    when "-v"; printVersion
    when "-h"; printHelp
    else
      # Error: To less and/or wrong argument(s):
      return "3"
    end
  end

  # Die ersten beiden Argumente müssen Dateien sein:
  for i in 0..1
    # Error: File does not exist:
    return "2" unless File.file?(args.at(i))
  end

  # returnvalue:
  return "0"  # alles OK
end




# ==============================================================================
# description  : Prints version
# requires     :
# returns      :
def printVersion()
  puts "Version: " + $VERSION
  exit
end




# ==============================================================================
# description  : Prints help
# requires     :
# returns      :
def printHelp()
  puts ""
  puts "-------------------------------------------------------"
  puts " bdiff - agregates a diff over two parts lists or generates"
  puts "         a total parts list over 2+ files."
  puts " by      :   " + $AUTHOR
  puts " version :   " + $VERSION
  puts ""
  puts " usage: bdiff.rb new old [-m] [-k] [-q] [-s]"
  puts "    new .... newer parts list (BOM)"
  puts "    old .... older parts list (BOM)"
  puts "    -mn .... operating mode (default = -m1):"
  puts "             0 = generate a total parts list over 2+ files"
  puts "             1 = aggregates increases compared to older BOM"
  puts "             2 = aggregates reductions compared to newer BOM"
  puts "    -kn .... number of column of ID key"
  puts "    -qn .... number of column of quantity"
  puts "    -s ..... separator as String"
  puts "-------------------------------------------------------"
  exit
end




# ==============================================================================
# description  : Initialisierung
# requires     :
# returns      : argList as Hash (Liste aller Parameter)
# args         : an das Programm übergebene Argumente (as Array)
def init(args)
  argList = Hash.new
  files = Array.new
  f = Pathname.new(args.at(0))

  # default Werte setzen:
  argList["mode"] = $MODE
  argList["posID"] = $POSID
  argList["posQty"] = $POSQTY
  argList["sep"] = $SEP

  # Auswertung und Übernahme der übergebenen Argumente:
  for i in 0..args.size-1
    # zwischenspeichern, ist übersichtlicher:
    arg = args[i]

    # Verarbeiten der Argumente:
    if File.file?(arg) then
      files[i] = arg
    else
      s = arg.slice(0,2)              # switch (-m..)
      v = arg.slice(2,arg.length-2)   # value (..12)
      case s
        when "-m"; argList["mode"] = v
        when "-k"; argList["posID"] = v
        when "-q"; argList["posQty"] = v
        when "-s"; argList["sep"] = v
      end
    end
  end

  argList["files"] = files  # files list

  # Dateiname der Zieldatei ermitteln (Postfix für Dateinamen abhängig
  # vom Modus):
  case argList["mode"]
    when "0"; pf = $POSTFIX0
    when "1"; pf = $POSTFIX1
    when "2"; pf = $POSTFIX2
  end
  argList["destPath"] = f.dirname
  argList["destFile"] = f.dirname.to_s + "/" +
    f.basename.sub(f.extname, "").to_s + pf + f.extname

  # returnvalue:
  argList
end




# ==============================================================================
# description  : Output of Error messages (and optional quit the script) if an
#                error occurs.
# requires     : -
# returns      : just an error prompt
# errCode      : Error Code (as String)
def outErr(errCode)
  case errCode
  when "1"; puts "Error: To less arguments."
  when "2"; puts "Error: File does not exist."
  when "3"; puts "Error: To less and/or wrong argument(s)."
  when "x"; puts "Error: undefined Error"
  else; puts "undefined Error"
  end

  # Programm beenden:
  exit
end




# ==============================================================================
# description  : Liest den Dateiinhalt aller aufgelisteten Dateien ein und gibt
# =>             sie als Array mit je einem Array je Datei zurück.
# requires     :
# returns      :
# files        : files list (full or relative pathes) as array
def getContent(files)
  content = Array.new
  for i in 0..files.size-1
    f = File.open(files[i], "r")
    content[i] = f.readlines
    f.close
    #puts "File " + i.to_s + " von " + (files.size).to_s + " gelesen."
  end
  # returnvalue:
  content
end




# ==============================================================================
# description  :
# requires     :
# returns      : Neuen Inhalt (Text) der Differenz-/Summenstückliste
# Modus 0 -- Summenstückliste:
# Modus 1 -- Diff (Mehrungen):
# Modus 2 -- Diff (Minderungen):
def diff(args, content)
  # Settings:
  newContent = ""
  files = args["files"]
  mode = args["mode"]
  posID = args["posID"].to_i
  posQty = args["posQty"].to_i
  sep = args["sep"]

  # Reihenfolge der Dateien in Abhängigkeit von Modus (1 oder 2) setzen:
  # Referenzdatei bei Mehrungen ist die Neuere, bei Minderungen die ältere
  # Übergeben werden soll immer die neuere als erstes Argument (einheitliche
  # Arbeitsweise).
  case mode
    when "1"; a = 0; b = 1 # a=neuer, b=älter
    when "2"; a = 1; b = 0 # a=älter, b=neuer
  end

  # Anzahl der Zeilen der zweiten Datei (Zum Erkennen des Listenendes im Array)
  # Notwendig für Artikel die in der alten Datei nicht enthalten sind.
  # Für diese kann keine Übereinstimmung gefunden werden:
  file2Size = content[b].size

  # Inhalt der ersten Datei durchsuchen:
  content[a].each { |line1|
    cntr = 0

    # ID und Menge des Artikels aus der ersten Liste:
    id1 = getCell(line1, posID, sep)
    qty1 = getCell(line1, posQty, sep)

    # Nur im Modus 0 (Summenstückliste):
    #TODO: noch fertigstellen

    # Nur im Modus 1 oder 2 ausführen (Diff):
    if (mode == "1") || (mode == "2") then
      # zweite Datei durchsuchen:
      content[b].each { |line2|
        # nur weiter machen wenn die Zeile die gesuchte 'ID' enthält:
        if line2.include?(id1) then
          # ID und Menge des Artikels aus der zweiten Liste:
          id2 = getCell(line2, posID, sep)
          qty2 = getCell(line2, posQty, sep)

          # Wenn 'ID' in der zweiten Liste gefunden wurde:
          if id1 == id2 then
            if qty2.to_i < qty1.to_i then
              diff = qty1.to_i - qty2.to_i

              # Ersetzen der alten Stückzahl mit der Differenz:
              line = line2.split(sep)
              line[posQty] = $TEXTQUAL + diff.to_s + $TEXTQUAL
              line = line.map { |i| i.to_s }.join(";")

              newContent = newContent + line
            end
          end

        else
          # Zähler hochzählen, bis das Ende der Vergleichsdatei erreicht wurde:
          cntr = cntr + 1

          # Ende der Vergleichsdatei erreicht, Artikel wurde nicht gefunden.
          # Deshalb: Die ganze Zeile an die Diff-Liste anhängen.
          if cntr >= file2Size then
            case mode
              when "1"; newContent = newContent + line1
              when "2"; newContent = newContent + line2
            end
          end

        end
      } # end each
    end
  }

  # returnvalue:
  newContent
end




# ==============================================================================
# description  :
# requires     :
# returns      :
# row          :
# col          :
# sep          :
def getCell(row, col, sep)
  return row.split(sep)[col].delete $TEXTQUAL + "\n"
end




# ==============================================================================
main()
# ==============================================================================




# ==============================================================================
#  Testszenario
#
=begin
* Stückzahl in neuerer Datei: erhöht, reduziert
* Artikel fehlt komplett in neuerer Datei
* Artikel fehlt komplett in älterer Datei
* Zusätzliche Leerzeilen
=end
